// This file is part of While, an educational programming language and program
// analysis framework.
//
//   Copyright 2023 Florian Brandner
//
// While is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// While is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// While. If not, see <https://www.gnu.org/licenses/>.
//
// Contact: florian.brandner@telecom-paris.fr
//

// A simple analysis determining whether symbolic registers contain a constant
// value.

#include "WhileAnalysis.h"
#include "WhileLang.h"
#include "WhileCFG.h"
#include "WhileColor.h"


enum WhileConstantKind
{
  TOP,
  BOTTOM,
  CONSTANT
};

struct WhileConstantValue
{
  WhileConstantKind Kind;
  int Value;

  WhileConstantValue() : Kind(TOP), Value(0)
  {
  }

  WhileConstantValue(int value) : Kind(CONSTANT), Value(value)
  {
  }

  WhileConstantValue(WhileConstantKind kind) : Kind(kind), Value(0)
  {
  }
};

bool operator==(const WhileConstantValue &a, const WhileConstantValue &b)
{
  return (a.Kind == b.Kind && a.Value == b.Value);
}

typedef std::map<int, WhileConstantValue> WhileConstantDomain;

std::ostream &operator<<(std::ostream &s, const WhileConstantValue &v)
{
  switch (v.Kind)
  {
    case TOP:
      return s << FLIGHT_GRAY << u8"⊤" << CRESET;
    case BOTTOM:
      return s << FRED << u8"⊥" << CRESET;
    case CONSTANT:
      return s << FGREEN << v.Value << CRESET;
  };
  abort();
}

struct WhileConstant : public WhileDataFlowAnalysis<WhileConstantDomain>
{
  std::ostream &dump_first(std::ostream &s,
                           const WhileConstantDomain &value) override
  {
    s << "    [";
    bool first = true;
    for(const auto&[idx, c] : value)
    {
      if (!first)
        s << ", ";

      s << "R" << idx << "=" << c;
      first = false;
    }
    return s << "]\n";
  }

  std::ostream &dump_pre(std::ostream &s,
                         const WhileConstantDomain &value) override
  {
    return s;
  }

  std::ostream &dump_post(std::ostream &s,
                          const WhileConstantDomain &value) override
  {
    return dump_first(s, value);
  }

  static void updateRegisterOperand(const WhileInstr &instr, unsigned int idx,
                             WhileConstantDomain &result,
                             WhileConstantValue value)
  {
    const WhileOperand &op = instr.Ops[idx];
    switch (op.Kind)
    {
      case WREGISTER:
        assert(op.ValueOrIndex >= 0);
        result[op.ValueOrIndex] = value;
        return;

      case WFRAMEPOINTER:
      case WIMMEDIATE:
      case WBLOCK:
      case WFUNCTION:
      case WUNKNOWN:
        assert("Operand is not a register.");
    }
    abort();
  }

  static WhileConstantValue readDataOperand(const WhileInstr &instr,
                                            unsigned int idx,
                                            const WhileConstantDomain &input)
  {
    const WhileOperand &op = instr.Ops[idx];
    switch (op.Kind)
    {
      case WREGISTER:
      {
        assert(op.ValueOrIndex >= 0);
        auto value = input.find(op.ValueOrIndex);
        if (value == input.end())
          return BOTTOM; // register undefined
        else
          return value->second;
      }
      case WIMMEDIATE:
        return op.ValueOrIndex;

      case WFRAMEPOINTER:
          return BOTTOM;

      case WBLOCK:
      case WFUNCTION:
      case WUNKNOWN:
        assert("Operand is not a data value.");
    }
    abort();
  }

  WhileConstantDomain transfer(const WhileInstr &instr, const WhileConstantDomain input) override
  {
    WhileConstantDomain result = input;
    const auto &ops = instr.Ops;
    switch(instr.Opc)
    {
      case WBRANCHZ:
      case WBRANCH:
      case WRETURN:
      case WSTORE:
        // do not write symbolic registers
        break;

      case WCALL:
      {
        // Ops: Fun Opd = Arg1, Arg2, ... ArgN
        //          ^ ... operand 1 is the destination register
        assert(ops.size() > 2);

        // Set the abstract value of operand 1, i.e., the destination register.
        updateRegisterOperand(instr, 1, result, 0 /* TODO: chose a value from the abstract domain, replacing 0 */);
        break;
      }

      case WLOAD:
      {
        // Ops: OpD = [BaseAddress + Offset]
        //      ^ ... operand 0 is the destination register
        assert(ops.size() ==  3);
        // TODO: Chose a value from the abstract domain and set the abstract value of operand 0, i.e., the destination register.
        // use updateRegisterOperand
        break;
      }

      case WPLUS:
      case WMINUS:
      case WMULT:
      case WDIV:
      case WEQUAL:
      case WUNEQUAL:
      case WLESS:
      case WLESSEQUAL:
      {
        // Ops: OpD = OpA + OpB
        //      ^     ^     ^ ... operand 2, a source register
        //      ^     ^ ... operand 1, a source register
        //      ^ ... operand 0, the destination register
        assert(ops.size() ==  3);
        WhileConstantValue a = readDataOperand(instr, 1, input);
        WhileConstantValue b = readDataOperand(instr, 2, input);

        if (a.Kind == CONSTANT && b.Kind == CONSTANT)
        {
          // Both values are known constants ...
          int aValue = a.Value;
          int bValue = b.Value;

          // Compute result using known constants
          int result;
          switch(instr.Opc)
          {
            case WBRANCHZ:
            case WBRANCH:
            case WRETURN:
            case WCALL:
            case WLOAD:
            case WSTORE:
                // impossible cases.
              abort();

            case WPLUS:
              // TODO: compute the result and assign a value to result.
              break;
            case WMINUS:
              // TODO: compute the result and assign a value to result.
              break;
            case WMULT:
              // TODO: compute the result and assign a value to result.
              break;
            case WDIV:
              // TODO: compute the result and assign a value to result.
              break;
            case WEQUAL:
              // TODO: compute the result and assign a value to result.
              break;
            case WUNEQUAL:
              // TODO: compute the result and assign a value to result.
              break;
            case WLESS:
              // TODO: compute the result and assign a value to result.
              break;
            case WLESSEQUAL:
              // TODO: compute the result and assign a value to result.
              break;
          };

          // TODO: Set the abstract value of the destination register using the result.
          // use updateRegisterOperand
        }
        else
        {
          // At least one of the values is not a constant ...
          // TODO: Chose a value from the abstract domain and set the abstract value of the destination register.
          // use updateRegisterOperand
        }
        break;
      }
    };

    return result;
  }

  static WhileConstantValue join(const WhileConstantValue &a,
                                 const WhileConstantValue &b)
  {
    // TODO: Implement the join function (replacing BOTTOM).
    return BOTTOM;
  }

  WhileConstantDomain join(std::list<WhileConstantDomain> inputs) override
  {
    WhileConstantDomain result;
    for(WhileConstantDomain r : inputs)
    {
      for(const auto&[idx, value] : r)
      {
        WhileConstantValue &resultvalue = result[idx];
        resultvalue = join(resultvalue, value);
      }
    }

    return result;
  }
};


struct WhileConstantRegisterAnalysis : public WhileAnalysis
{
  void analyze(const WhileProgram &p) override
  {
    WhileConstant WCRA;
    WCRA.analyze(p);
    WCRA.dump(std::cout, p);
  };

  WhileConstantRegisterAnalysis() : WhileAnalysis("WCRA",
                                                  "Constant Register Analysis")
  {
  }
};

WhileConstantRegisterAnalysis WCRA;
