// This file is part of While, an educational programming language and program
// analysis framework.
//
//   Copyright 2023 Florian Brandner
//
// While is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// While is distributed isn the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// While. If not, see <https://www.gnu.org/licenses/>.
//
// Contact: florian.brandner@telecom-paris.fr
//

// A simple inter-procedural analysis tracking all possible values that the
// FramePointer might have at any location in the program.

#include "WhileAnalysis.h"
#include "WhileLang.h"
#include "WhileCFG.h"
#include "WhileColor.h"

#include <array>

enum Stages
{
  IF,
  ID,
  EX,
  MEM,
  WB,
};

typedef std::array<const WhileInstr*, WB> WhilePipelineState;

typedef std::set<WhilePipelineState> WhilePipelineStates;

struct WhilePipelineAnalysis
  : public WhileInterproceduralDataFlowAnalysis<WhilePipelineStates>
{
  static std::ostream &dump_state(std::ostream &s, const WhilePipelineState &ps)
  {
    s << "{";
    bool first = true;
    for(const WhileInstr *i : ps)
    {
      if (!first)
        s << ", ";

      if (i)
        i->dumpid(s);
      else
        s << "(NOP)";

      first = false;
    }
    return s << "}";
  }

  std::ostream &dump_entry(std::ostream &s,
                           const WhilePipelineStates &value) override
  {
    s << "[";
    bool first = true;
    for(const WhilePipelineState &ps : value)
    {
      if (!first)
        s << ", ";

      dump_state(s, ps);
      first = false;
    }
    return s << "]";
  }

  std::ostream &dump_first(std::ostream &s,
                           const WhilePipelineStates &value) override
  {
    s << "    ";
    return dump_entry(s, value) << "\n";
  }

  std::ostream &dump_pre(std::ostream &s,
                         const WhilePipelineStates &value) override
  {
    return s;
  }

  std::ostream &dump_post(std::ostream &s,
                          const WhilePipelineStates &value) override
  {
    return dump_first(s, value);
  }

  virtual WhilePipelineStates initialize(const WhileFunction *f) override
  {
    WhilePipelineStates result;
    if (f->CallSites.empty())
      result.emplace(WhilePipelineState());

    return result;
  }

  WhilePipelineStates transfer(const WhileInstr &instr,
                               const WhilePipelineStates input) override
  {
    WhilePipelineStates result;
    for(const WhilePipelineState &ps : input)
    {
      WhilePipelineState tmp;

      // The instruction currently fetched enters the pipeline.
      tmp[IF] = &instr;

      // The other instructions move to next pipeline stage.
      auto i(ps.begin());
      for(unsigned int s = ID; s < WB; s++)
        tmp[s] = *i++;

      result.emplace(tmp);
    }

    return result;
  }

  WhilePipelineStates join(std::list<WhilePipelineStates> inputs) override
  {
    WhilePipelineStates result;
    for(const WhilePipelineStates &v : inputs)
      result.insert(v.begin(), v.end());

    return result;
  }
};


struct WhileInterproceduralPipelineAnalysis : public WhileAnalysis
{
  void analyze(const WhileProgram &p) override
  {
    WhilePipelineAnalysis WIPA;
    WIPA.analyze(p);
    WIPA.dump(std::cout, p);
  };

  WhileInterproceduralPipelineAnalysis() : WhileAnalysis("WIPA",
                                           "Inter-procedural Pipeline Analysis")
  {
  }
};

WhileInterproceduralPipelineAnalysis WIPA;
